/*
 * selector.c
 *
 * select a small group of pictures from a large directory tree and
 * send them to the Ceiva picture frame, one a time, updating the display
 * control file.
 *
 * notices time of day and blanks display
 *
 * Brad Parker <brad@heeltoe.com>
 *
 * $Id: selector.c,v 1.5 2005/09/17 13:56:37 brad Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <time.h>

#include <unistd.h>
#include <dirent.h>

#include <sys/types.h>
#include <sys/stat.h>

#define MAX_DIRNAMES	10
char *dirnames[MAX_DIRNAMES];
int dirname_count;

int need_dir_scan;
int debug_dir;
int sleep_time;
int control_delay;
int do_system;
int do_send;
int do_verbose;

int blank_hour_on;
int blank_hour_off;

char *tmp_prefix = "/ram";
char *control_file_name = "control.new";
char *data_file_name = "file";
char *control_file_path;

struct file_entry {
  int transformed;
  int valid;
  int d_type;
  char name[1];
};

struct file_entry **file_vector;
struct file_entry **other_vector;
int file_vector_size;
int file_vector_max;
#define FILE_VECTOR_MAX	1 /* 1024 */


struct file_entry *picks[10];
int num_picks, max_picks, base_pick;

//char convert_command[] = "/usr/X11R6/bin/convert -gamma 1.2 -scale 640x480 -quality 80 -colors 4096 %s /ram/%s";
char convert_command[] = "/usr/X11R6/bin/convert -gamma 1.2 -scale 640x480 -quality %d %s %s %s";

char send_command[] = "/usr/local/bin/sz -f %s %s";

char move_command[] = "/usr/local/bin/zcommand %s 'mv /ram/control.new /ram/control'";

char blank_command[] = "/usr/local/bin/zcommand %s '/usr/local/bin/backlight 4095'";

//char unblank_command[] = "/usr/local/bin/zcommand %s '/usr/local/bin/backlight 2560'";
char unblank_command[] = "/usr/local/bin/zcommand %s '/usr/local/bin/backlight 2816'";

char size_command[] = "/usr/bin/rdjpgcom -verbose %s";

char send_command_options[] = "-D /dev/ttyS2 -s 38400";

int
select_init(void)
{
  num_picks = 0;
  max_picks = 5;
  base_pick = 0;
}

int
get_jpeg_size(char *filename, int *px, int *py)
{
  FILE *f;
  int w, h, n;
  char cmd[1024];

  sprintf(cmd, size_command, filename);
  if (do_verbose) printf("cmd: '%s'\n", cmd);

  f = popen(cmd, "r");

  if (f == NULL)
    return -1;

  if (f) {
    /* JPEG image is 160w * 120h, 3 color components, 8 bits per sample */
    w = h = 0;
    n = fscanf(f, "JPEG image is %dw * %dh", &w, &h);
    *px = w;
    *py = h;

    //printf("n=%d '%s' w=%d h=%d\n", n, filename, w, h);
  }

  pclose(f);

  return 0;
}

int
copy_file(char *oldfilename, char *newfilename, int *psize)
{
  char cmd[1024];

  sprintf(cmd, "cp %s %s", oldfilename, newfilename);
  if (do_verbose)
    printf("cmd: %s\n", cmd);

  *psize = 0;

  if (do_system) {
    struct stat s;

    if (stat(oldfilename, &s)) {
      perror(oldfilename);
      return -1;
    }

    *psize = s.st_size;

    if (do_verbose) printf("size %d\n", s.st_size);

    if (s.st_size > 100000)
      return -1;

    system(cmd);

    return 0;
  }

  return 0;
}

int
convert_file(int quality, char *oldfilename, char *newfilename, int *psize)
{
  char *other;
  char cmd[1024];

  if (0)
    printf("convert_file(quality=%d, oldfilename='%s', newfilename='%s'\n",
	   quality, oldfilename, newfilename);

  other = "";
  //  other = "-colors 1024";
  //  other = "-colors 512 -dither";
  //  other = "-colors 512 -treedepth 4";
  //other = "-colors 256 -treedepth 6";

  sprintf(cmd, convert_command, quality, other, oldfilename, newfilename);
  if (do_verbose)
    printf("cmd: %s\n", cmd);

  *psize = 0;

  if (do_system) {
    struct stat s;

    system(cmd);

    if (stat(newfilename, &s)) {
      perror(newfilename);
      return -1;
    }

    *psize = s.st_size;

    if (do_verbose) printf("quality %d, size %d\n", quality, s.st_size);

    return 0;
  }

  return 0;
}

int
get_send_cmd_options(char **ptr)
{
  static char options[256];
  strcpy(options, send_command_options);
  *ptr = options;
  return 0;
}

int
select_transform_file(char *oldfilename, char *newfilename)
{
  int size, quality, x, y;
  char cmd[1024], *options;

  if (0) printf("select_transform(oldfilename='%s', newfilename='%s'\n",
		oldfilename, newfilename);

  if (get_jpeg_size(oldfilename, &x, &y))
    return -1;

  if (x > 640 || y > 480) {

shrink:
    quality = 80;
    if (convert_file(quality, oldfilename, newfilename, &size))
      return -1;

    while (size > 100000 && quality > 0) {
      quality -= 10;
      if (convert_file(quality, oldfilename, newfilename, &size))
	return -1;
    }

  } else {
    if (copy_file(oldfilename, newfilename, &size))
      return -1;

    if (size > 100000)
      goto shrink;
  }

  /* set download speed */
  get_send_cmd_options(&options);

  if (size <= 100000) {
    sprintf(cmd, send_command, options, newfilename);

    if (do_verbose) printf("cmd: %s\n", cmd);

    if (do_system && do_send) system(cmd);
  }

#if 0
  /* create new control file */
  sprintf(cmd, send_command, options, control_file_path);

  if (do_verbose) printf("cmd: %s\n", cmd);

  if (do_system && do_send) system(cmd);

  /* and, do atomic move */
  sprintf(cmd, move_command, options, control_file_path);

  if (do_verbose) printf("cmd: %s\n", cmd);

  if (do_system && do_send) system(cmd);
#endif

  return 0;
}

int
blank_display(void)
{
  char *options;
  char cmd[1024];

  get_send_cmd_options(&options);

  sprintf(cmd, blank_command, options);

  if (do_verbose) printf("cmd: %s\n", cmd);

  if (do_system && do_send) system(cmd);

  return 0;
}

int
unblank_display(void)
{
  char *options;
  char cmd[1024];

  get_send_cmd_options(&options);

  sprintf(cmd, unblank_command, options);

  if (do_verbose) printf("cmd: %s\n", cmd);

  if (do_system && do_send) system(cmd);

  return 0;
}

int
select_transform(void)
{
  int i, newfilenum;
  char newfilename[1024];

  for (i = 0; i < num_picks; i++) {
    if (!picks[i]->transformed) {

      if (num_picks < max_picks)
	newfilenum = num_picks-1;
      else
	newfilenum = (base_pick + max_picks - 1) % max_picks;

      sprintf(newfilename, "%s/%s%d", tmp_prefix, data_file_name, newfilenum);

      if (select_transform_file(picks[i]->name, newfilename) == 0) {
	picks[i]->valid = 1;
	picks[i]->transformed = 1;
      }
    }
  }

  return 0;
}

/*
 * build and send control file
 *
 * it's important that we send a sequence of control files which
 * display a file, then don't display it when it's about to be updated;
 * otherwise the frame will be displaying an image we're updating.
 *
 */
int
select_control(void)
{
  int i, count, base;
  char newfilename[1024], line[128];
  FILE *f;
  static int oddfill;

  f = fopen(control_file_path, "w");
  if (f == NULL)
    return -1;

  if (!do_send) {
    printf("base_pick %d\n", base_pick);
    printf("%s\n", control_file_path);
  }

  if (num_picks < max_picks) {
    count =  num_picks;
    base = 0;
  } else {
    count = num_picks-1;
    base = base_pick+1;
  }

  for (i = 0; i < count; i++) {
    sprintf(newfilename, "%s%d", data_file_name, (base+i) % max_picks);

    sprintf(line, "%s/%s %d%s",
	    tmp_prefix, newfilename, control_delay,
	    oddfill ? " " : "");
    if (oddfill) oddfill = 0; else oddfill = 1;

    fprintf(f, "%s\n", line);

    if (!do_send || do_verbose) printf("line%d: %s\n", i+1, line);
  }

  fclose(f);

#if 1
  {
  char cmd[1024], *options;

  get_send_cmd_options(&options);

  /* create new control file */
  sprintf(cmd, send_command, options, control_file_path);

  if (do_verbose) printf("cmd: %s\n", cmd);

  if (do_system && do_send) system(cmd);

  /* and, do atomic move */
  sprintf(cmd, move_command, options, control_file_path);

  if (do_verbose) printf("cmd: %s\n", cmd);

  if (do_system && do_send) system(cmd);
  }
#endif

  return 0;
}


int
select_send(void)
{
  select_transform();
  select_control();

  if (num_picks == max_picks) {
    base_pick = (base_pick + 1) % max_picks;
  }

  return 0;
}

int
select_pick(struct file_entry *fe)
{
  int i;

  fe->transformed = 0;
  fe->valid = 0;

  /* if 'picks' vector is full, shuffle down one and add to end */
  if (num_picks == max_picks) {
    for (i = 0; i < max_picks; i++) {
      picks[i] = picks[i+1];
    }

    picks[max_picks-1] = fe;
    return 0;
  }

  /* 'picks' vector is not full, augment */
  picks[num_picks] = fe;
  num_picks++;

  return 0;
}

int num_times_picked;
int other_pick_count = 4;

int
select_next(void)
{
  double n;
  long r;
  int pick;

  if (other_vector && ++num_times_picked > other_pick_count) {
    num_times_picked = 0;
    select_pick(other_vector[0]);
    return;
  }

  r = random();
  n = (double)r / RAND_MAX;

  pick = file_vector_size * n;

  select_pick(file_vector[pick]);
}

/* */
int
dir_ok(char *dname)
{
  if (strcasecmp(dname, "Thumbnails") == 0)
    return 0;

  return 1;
}

/* */
int 
path_ok(char *path)
{
  /* no path names with spaces in them */
  if (strchr(path, ' '))
    return 0;

  return 1;
}

/* file extension we want to collect */
int
extension_ok(char *fname)
{
  int l = strlen(fname);

  /* ignore "index*" */
  if (memcmp(fname, "index", 5) == 0)
    return 0;

  /* no file names with spaces in them */
  if (strchr(fname, ' '))
    return 0;

  /* allow only proper extensions */
  if (strncasecmp(fname + l - 4, ".jpg", 4) == 0)
    return 1;

  if (strncasecmp(fname + l - 4, ".gif", 4) == 0)
    return 1;

  return 0;
}

void
file_vector_dump(void)
{
  int i;

  printf("file_vector %p, file_vector_size %d, file_vector_max %d\n",
	 file_vector, file_vector_size, file_vector_max);

  for (i = 0; i < file_vector_size; i++) {
    printf("[%d] %x '%s'\n",
	   i, file_vector[i]->d_type, file_vector[i]->name);
  }
}

void
file_vector_free(void)
{
  int i;

  for (i = 0; i < file_vector_size; i++) {
    free(file_vector[i]);
    file_vector[i] = 0;
  }

  file_vector_size = 0;
  file_vector_max = 0;

  if (file_vector) {
    free(file_vector);
    file_vector = 0;
  }
}

int
file_vector_add(struct file_entry *fe)
{
  struct file_entry **new;
  int old_file_vector_max;

  if (file_vector_size >= file_vector_max) {

    /* remember old size and then double */
    old_file_vector_max = file_vector_max;

    if (file_vector_max == 0)
      file_vector_max = FILE_VECTOR_MAX;
    else
      file_vector_max *= 2;

    if (0) printf("file_vector_max %d\n", file_vector_max);

    new = (struct file_entry **)malloc(sizeof(struct file_entry *) *
				       file_vector_max);
    if (new == NULL)
      return -1;

    /* if we had an old vector, copy to new and free it */
    if (old_file_vector_max) {
      memcpy(new, file_vector,
	     sizeof(struct file_entry *) * old_file_vector_max);

      free(file_vector);
    }

    file_vector = new;
  }

  file_vector[file_vector_size++] = fe;

  return 0;
}

int
file_add(char *dirname, int d_type, char *d_name)
{
  struct file_entry *fe;

  if (debug_dir)
    printf("file_add(dirname='%s', d_type=%x, d_name='%s')\n",
	   dirname, d_type, d_name);

  fe = (struct file_entry *)malloc(sizeof(struct file_entry) +
				   strlen(dirname) + strlen(d_name));
  if (fe == NULL)
    return -1;

  fe->d_type = d_type;
  strcpy(fe->name, dirname);
  strcat(fe->name, "/");
  strcat(fe->name, d_name);

  if (file_vector_add(fe)) {
    free(fe);
    return -1;
  }

  return 0;
}


int
dir_add(char *dirname)
{
  char *dir;

  if (debug_dir) printf("dir_add(dirname='%s')\n", dirname);

  if (dirname_count >= MAX_DIRNAMES)
    return -1;

  if (!dirname[0])
    return -1;

  dir = strdup(dirname);

  if (dir[strlen(dir)-1] == '/') {
    dir[strlen(dir)-1] = '\0';
  }

  dirnames[dirname_count++] = dir;

  if (debug_dir)
    printf("dir_add(dirname='%s') dirname_count %d\n", dir, dirname_count);

  return 0;
}

int
dir_scan(char *dirname)
{
  DIR *d;
  struct dirent *de;
  struct stat s;
  char name[1024];

  if (debug_dir) printf("dir_scan(dirname='%s')\n", dirname);

  d = opendir(dirname);
  if (d != NULL) {

    if (debug_dir) printf("d %p\n", d);

    while ((de = readdir(d)) != NULL) {

      if (debug_dir) printf("de %p, '%s'\n", de, de->d_name);

      if (de->d_name[0] == '.')
	continue;

      strcpy(name, dirname);
      strcat(name, "/");
      strcat(name, de->d_name);

      if (de->d_type == 0) {
	stat(name, &s);

	if (S_ISREG(s.st_mode))
	  de->d_type = DT_REG;

	if (S_ISDIR(s.st_mode))
	  de->d_type = DT_DIR;
      }

      switch (de->d_type) {
      case DT_DIR:
	if (debug_dir) printf("dir: %s\n", name);

	if (dir_ok(de->d_name)) {
	  dir_scan(name);
	}
	break;
      case DT_REG:
      case DT_LNK:
	if (debug_dir) printf("file: %s\n", name);
	if (extension_ok(de->d_name) && path_ok(dirname)) {
	  file_add(dirname, de->d_type, de->d_name);
	}
	break;
      default:
      }
    }

    closedir(d);
  }
}

int
dir_scan_all(void)
{
  int i;

  if (debug_dir) printf("dir_scan_all()\n");

  file_vector_free();
  select_init();

  for (i = 0; i < dirname_count; i++) {
    dir_scan(dirnames[i]);
  }

  if (do_verbose) printf("collected %d filenames\n", file_vector_size);

  return 0;
}

int
setup_other_vector(char *dirname, char *d_name)
{
  struct file_entry *fe;

  fe = (struct file_entry *)malloc(sizeof(struct file_entry) +
				   strlen(dirname) + strlen(d_name));

  fe->d_type = DT_REG;
  strcpy(fe->name, dirname);
  strcat(fe->name, "/");
  strcat(fe->name, d_name);

  other_vector = (struct file_entry **)malloc(sizeof(struct file_entry *));

  other_vector[0] = fe;

  return 0;
}

void
setup_paths(void)
{
  char path[1024];

  sprintf(path, "%s/%s", tmp_prefix, control_file_name);

  control_file_path = strdup(path);

  if (do_verbose) printf("control file: '%s'\n", control_file_path);
}


void setup_signals(void);

void
sighup_handler(int arg)
{
  need_dir_scan++;
  setup_signals();
}

void
setup_signals(void)
{
  signal(SIGHUP, sighup_handler);
}

char get_command[] = "cd /tmp; rm -f KBOSMOS.gif; wget http://www.wunderground.com/data/images/KBOSMOS.gif";
char cvt_command[] = "cd /tmp; giftopnm KBOSMOS.gif | ppmtojpeg >weather.jpg";

void
hourly_event(void)
{
  if (do_verbose) {
    printf("hourly event\n");
    printf("cmd: %s\n", get_command);
  }

  system(get_command);

  if (do_verbose)
    printf("cmd: %s\n", cvt_command);

  system(cvt_command);
}

static int blanked;
int
blank_event(void)
{
  if (blanked)
    return;
  blank_display();
  blanked = 1;
}

int
unblank_event(void)
{
  if (!blanked)
    return;
  unblank_display();
  blanked = 0;
}

static int last_hour;

void
check_daytime(void)
{
  struct timeval tv;
  struct tm *tm;
  time_t t;

  gettimeofday(&tv, 0);
  t = time(0);
  tm = localtime(&t);

 if (tm->tm_hour >= blank_hour_on || tm->tm_hour <= blank_hour_off) {
    blank_event();
 } else {
    unblank_event();
 }

//last_hour = 0; /* debug */
  if (tm->tm_hour != last_hour) {
    last_hour = tm->tm_hour;
    hourly_event();
  }
}

void
run(void)
{
  if (do_verbose) printf("run()\n");

  setup_signals();

  need_dir_scan = 1;

  unblank_display();

  while (1) {
    check_daytime();
    
    if (need_dir_scan) {
      need_dir_scan = 0;
      dir_scan_all();
      if (0) file_vector_dump();
    }

    if (!blanked) {
      if (file_vector_size > 0) {
	select_next();
	select_send();
      }
    }

    if (do_system) sleep(sleep_time);
  }
}

void
usage(const char *argv0)
{
  fprintf(stderr, "usage:\n");
  fprintf(stderr, "%s {options} <path-to-picture-tree>\n", argv0);
  fprintf(stderr, "-c <seconds>   set control delay\n");
  fprintf(stderr, "-d             debug\n");
  fprintf(stderr, "-r             set random seed\n");
  fprintf(stderr, "-s <seconds>   set sleep time\n");
  fprintf(stderr, "-v             set verbose mode\n");
  fprintf(stderr, "-D             debug, don't send, don't execute commands\n");
  exit(1);
}

extern char *optarg;
extern int optind, opterr, optopt;

int
main(int argc, char *argv[])
{
  int c;

#define DEF_BLANK_HOUR_OFF	12+10 	/* 10pm */
#define DEF_BLANK_HOUR_ON	6 	/* 6am */

#define DEF_SLEEP_TIME		3*6
#define DEF_CONTROL_DELAY	50
#define MIN_CONTROL_DELAY	10

  sleep_time = DEF_SLEEP_TIME;
  control_delay = DEF_CONTROL_DELAY;
  do_system = 1;
  do_send = 1;

  blank_hour_on = DEF_BLANK_HOUR_OFF;
  blank_hour_off = DEF_BLANK_HOUR_ON;

  if (argc <= 1)
    usage(argv[0]);

  while ((c = getopt(argc, argv, "c:dDs:rvb:")) != -1) {
    switch (c) {
    case 'b':
      blank_hour_off = atoi(optarg);
      if (blank_hour_off < 0 || blank_hour_off > 23)
	blank_hour_off = DEF_BLANK_HOUR_OFF;
      break;
    case 'c':
      control_delay = atoi(optarg);
      if (control_delay <= MIN_CONTROL_DELAY)
	control_delay = MIN_CONTROL_DELAY;
      break;
    case 'd':
      debug_dir = 1;
      break;
    case 'r':
      srandom(time(0L));
      break;
    case 's':
      sleep_time = atoi(optarg);
      if (sleep_time <= 0)
	sleep_time = 1;
      break;
    case 'v':
      do_verbose = 1;
      break;
    case 'D':
      do_system = 0;
      do_send = 0;
      do_verbose = 1;
      break;
    default:
#if 0
      printf("c %d\n", c);
      break;
#endif
      usage(argv[0]);
    }
  }

  //printf("optind %d, argc %d\n", optind, argc);

  while (optind < argc) {
    dir_add(argv[optind++]);
  }

  setup_paths();
  setup_other_vector("/tmp", "weather.jpg");

  run();

  exit(0);
}
